
import json

import spacy
import numpy as np
from nltk.tokenize import sent_tokenize
from nltk.corpus import stopwords

class SemanticSimilarityTextChunker:
    def __init__(self, query):
        self.query = query
        self.nlp = spacy.load("en_core_web_sm")
        self.stop_words = set(stopwords.words('english'))

    def preprocess_text(self, text):
        doc = self.nlp(text.lower())
        return [token.lemma_ for token in doc if token.text not in self.stop_words and token.is_alpha]

    def get_query_keywords(self):
        return self.preprocess_text(self.query)

    def calculate_similarity(self, word1, word2):
        return word1.similarity(word2)

    def score_sentence(self, sentence, keywords):
        preprocessed_sentence = self.preprocess_text(sentence)
        doc_sentence = self.nlp(" ".join(preprocessed_sentence))
        doc_keywords = self.nlp(" ".join(keywords))

        if not doc_sentence.vector.any() or not doc_keywords.vector.any():
            return 0

        # Calculate the average similarity between each word in the sentence and the keywords
        similarities = []
        for token in doc_sentence:
            if token.has_vector:
                keyword_similarities = [self.calculate_similarity(token, keyword) for keyword in doc_keywords if
                                        keyword.has_vector]
                if keyword_similarities:
                    similarities.append(max(keyword_similarities))

        # Calculate the average similarity score
        avg_similarity = np.mean(similarities) if similarities else 0

        # Adjust the score based on the presence of exact matches
        exact_match_bonus = sum(1 for word in preprocessed_sentence if word in keywords)
        adjusted_score = (avg_similarity + 0.1 * exact_match_bonus) / (1 + 0.1 * len(keywords))

        return adjusted_score

    def chunk_text(self, text, chunk_size=3,top_chunks_in_article=3):
        sentences = sent_tokenize(text)
        keywords = self.get_query_keywords()

        # Score each sentence
        scored_sentences = [(sentence, self.score_sentence(sentence, keywords)) for sentence in sentences]

        # Sort sentences by score in descending order
        scored_sentences.sort(key=lambda x: x[1], reverse=True)

        # Group top-scoring sentences into chunks
        chunks = []
        for i in range(0, min(len(scored_sentences), chunk_size * 3), chunk_size):
            chunk = ' '.join([sent[0] for sent in scored_sentences[i:i + chunk_size]])
            chunks.append(chunk)

        return chunks[:top_chunks_in_article]  # Return top 3 chunks


from .researcher_ai import ContextExtractor,MultiAgentQueryOrchestrator
from .langchain_ollama import CustomChatOllama
from .web_search import WebSearchAgent,FastHTMLParserV3,OptimizedMultiQuerySearcher,WebSearchArguments
from .researcher_ai.main.search_provider.searcher import RealTimeGoogleSearchProvider
from .researcher_ai.main.parse_url.html_parser import UrlTextParser


import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count
from urllib.parse import quote_plus

from nodriver.cdp.dom import query_selector
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pydantic import BaseModel
from typing import List

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

class SearchResult(BaseModel):
    query: str
    title_and_links: list = []
    urls: list = []
    search_provider: str  = ""

class OptimizedMultiQuerySearcher:
    def __init__(self, chromedriver_path="/usr/local/bin/chromedriver", max_workers=None):
        self.chromedriver_path = chromedriver_path
        self.max_workers = max_workers or min(32, cpu_count() -2)
        self.driver_pool = []
        self.params = {
            "bing":{"search":"b_results","find_element":".b_algo","head_selector":"h2"},
            "google": {"search": "search", "find_element": "div.g", "head_selector": "h3"}
        }

    def _setup_driver(self):
        chrome_options = Options()
        chrome_options.add_argument("--headless")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--no-sandbox")
        chrome_options.add_argument("--disable-dev-shm-usage")
        chrome_options.add_argument("--disable-extensions")
        chrome_options.add_argument("--disable-images")
        chrome_options.add_argument("--blink-settings=imagesEnabled=false")
        chrome_options.add_argument("--start-maximized")
        chrome_options.page_load_strategy = 'none'
        return webdriver.Chrome(self.chromedriver_path, options=chrome_options)

    def _get_driver(self):
        if not self.driver_pool:
            return self._setup_driver()
        return self.driver_pool.pop()

    def _return_driver(self, driver):
        self.driver_pool.append(driver)

    def cleanup(self):
        if hasattr(self, 'driver'):
            self.driver.quit()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def search_single_query(self, query, num_results=10,
                            search_provider="google"
                            ):
        driver = self._get_driver()
        try:
            encoded_query = quote_plus(query)
            search_url = f"https://www.{search_provider}.com/search?q={encoded_query}"
            driver.get(search_url)

            WebDriverWait(driver, 5).until(
                EC.presence_of_element_located((By.ID, self.params[search_provider]["search"]))
            )
            # results = driver.find_elements(By.CSS_SELECTOR, self.params[search_provider]["find_element"])

            # search_results = []
            # for result in results[:num_results]:
            #     try:
            #         title = result.find_element(By.CSS_SELECTOR, "h3").text
            #         link = result.find_element(By.CSS_SELECTOR, "a").get_attribute("href")
            #         search_results.append({"title": title, "link": link})
            #     except Exception as e:
            #         logger.error(f"Error parsing result: {str(e)}")
            #         continue

            script = """
                        var results = [];
                        var elements = document.querySelectorAll('##find_element##');
                        for (var i = 0; i < arguments[0] && i < elements.length; i++) {
                            var element = elements[i];
                            var titleElement = element.querySelector('##head_selector##');
                            var linkElement = element.querySelector('a');
                            if (titleElement && linkElement) {
                                results.push({
                                    title: titleElement.innerText,
                                    link: linkElement.href
                                });
                            }
                        }
                        return results;
                        """
            script = script.replace("##find_element##",self.params[search_provider]["find_element"])
            script = script.replace("##head_selector##",self.params[search_provider]["head_selector"])
            search_results = driver.execute_script(script, num_results)

            urls = [x['link'] for x in search_results if len(x['link'])>0]

            return SearchResult(
                query=query,
                title_and_links=search_results,
                urls=urls,
                search_provider = search_provider
            )
        except Exception as e:
            logger.error(f"An error occurred while searching '{query}': {str(e)}")
            return SearchResult(query=query)
        finally:
            self._return_driver(driver)

    async def search_multiple_queries(self, queries: List[str], num_results=10,
                                      search_provider=None,return_only_urls = False
                                      ) -> List[SearchResult]:
        loop = asyncio.get_event_loop()
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            providers_list = self.params.keys() if search_provider is None else [search_provider]
            tasks= [
                loop.run_in_executor(
                    executor,
                    self.search_single_query,
                    query,
                    num_results,
                    provider
                )
                for query in queries for provider in providers_list
            ]
            result = await asyncio.gather(*tasks)
            if return_only_urls:
                all_urls = [url.urls for url in result]
                filtered_urls = [y for x in zip(*all_urls) for y in x]
                return filtered_urls
            return result

    def __del__(self):
        for driver in self.driver_pool:
            driver.quit()

if __name__ == '__main__':
    import time

    async def main():
        start_time = time.perf_counter()
        gs = OptimizedMultiQuerySearcher()
        # queries = ["who is narendra modi?", "what is python programming?", "latest news in technology"]
        queries=['Latest advancements in agentbased RAG models',
                  'RAG based approaches for autonomous agents recent research and applications',
                  'Comparative analysis of current agentbased RAG systems for specific application domain',
                  'Exploring the future of agentbased language understanding with RAG',
                  'OpenAIs AGI Evaluating its potential within a RAG framework'
                 ]

        results = await gs.search_multiple_queries(queries,search_provider="bing")
        end_time = time.perf_counter()
        for result in results:
            print(f"Query: {result.query}")
            print(f"Number of results: {len(result.title_and_links)}")
            print("---")
        print(f"Total time taken : {end_time-start_time}s")
        print(results)

    asyncio.run(main())


from typing import Any, Dict, List, Optional, Union, Tuple

import aiohttp
from langchain_core.callbacks import CallbackManagerForLLMRun
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import AIMessage, BaseMessage, ChatMessage, HumanMessage, SystemMessage, AIMessageChunk
from langchain_core.outputs import ChatGeneration, ChatGenerationChunk, ChatResult
import requests
import json
import asyncio

class CustomChatOllama(BaseChatModel):
    model: str = "qwen2:7b-instruct"
    base_url: str = "http://192.168.162.49:8888"
    temperature: float = 0.0
    num_predict=-1

    @property
    def _llm_type(self) -> str:
        return "custom-ollama-chat"

    def _convert_messages_to_prompt(self, messages: List[BaseMessage]) -> str:
        prompt = ""
        for message in messages:
            if isinstance(message, SystemMessage):
                prompt += f"System: {message.content}\n"
            elif isinstance(message, HumanMessage):
                prompt += f"Human: {message.content}\n"
            elif isinstance(message, AIMessage):
                prompt += f"AI: {message.content}\n"
            elif isinstance(message, ChatMessage):
                prompt += f"{message.role.capitalize()}: {message.content}\n"
        return prompt.strip()

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        prompt = self._convert_messages_to_prompt(messages)
        response = requests.post(
            f"{self.base_url}/chat",
            json={
                "text": prompt,
                "model": self.model,
                "temperature": self.temperature,
                "num_predict": kwargs.get("num_predict", 2048)
            },
            stream=True
        )

        full_response = ""
        for chunk in response.iter_content(chunk_size=None, decode_unicode=True):
            if chunk:
                full_response += chunk
                if run_manager:
                    run_manager.on_llm_new_token(chunk)

        chat_generation = ChatGeneration(
            message=AIMessage(content=full_response),
            generation_info={"finish_reason": "stop"}
        )
        return ChatResult(generations=[chat_generation])

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Union[ChatGenerationChunk, List[ChatGenerationChunk]]:
        prompt = self._convert_messages_to_prompt(messages)
        response = requests.post(
            f"{self.base_url}/chat",
            json={
                "text": prompt,
                "model": self.model,
                "temperature": self.temperature,
                "num_predict": kwargs.get("num_predict", 2048)
            },
            stream=True
        )

        for chunk in response.iter_content(chunk_size=None, decode_unicode=True):
            if chunk:
                chat_chunk = ChatGenerationChunk(
                    message=AIMessageChunk(content=chunk),
                    generation_info=None
                )
                if run_manager:
                    run_manager.on_llm_new_token(chunk)
                yield chat_chunk

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        prompt = self._convert_messages_to_prompt(messages)
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{self.base_url}/chat",
                json={
                    "text": prompt,
                    "model": self.model,
                    "temperature": self.temperature,
                    "num_predict": kwargs.get("num_predict", 2048)
                }
            ) as response:
                full_response = await response.text()

        chat_generation = ChatGeneration(
            message=AIMessage(content=full_response),
            generation_info={"finish_reason": "stop"}
        )
        return ChatResult(generations=[chat_generation])

    async def _abatch(
        self,
        messages: List[List[BaseMessage]],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> List[ChatResult]:
        prompts = [self._convert_messages_to_prompt(msg) for msg in messages]
        async with aiohttp.ClientSession() as session:
            tasks = [
                session.post(
                    f"{self.base_url}/chat",
                    json={
                        "text": prompt,
                        "model": self.model,
                        "temperature": self.temperature,
                        "num_predict": kwargs.get("num_predict", 2048)
                    }
                )
                for prompt in prompts
            ]
            responses = await asyncio.gather(*tasks)
            results = []
            for response in responses:
                full_response = await response.text()
                chat_generation = ChatGeneration(
                    message=AIMessage(content=full_response),
                    generation_info={"finish_reason": "stop"}
                )
                results.append(ChatResult(generations=[chat_generation]))
        return results


import asyncio
import time

from crewai import Agent, Task, Crew
from langchain.tools import Tool
from langchain_core.language_models import BaseChatModel

from .agent import WebSearchAgent, OptimizedMultiQuerySearcher
from .html_parser import FastHTMLParserV3
from ..langchain_ollama import CustomChatOllama
from .text_filter import TextFilter



class CrewAIWebSearchQASystem:
    def __init__(self,llm:BaseChatModel):
        self._llm = llm
        self._search_tool = None
        self._researcher = None
        self._setup_tool_and_search()


    def _web_search(self,query: str) -> str:
        """Run query through GoogleSearch and parse result."""
        web_agent = WebSearchAgent()
        rq = web_agent.refine_queries(user_query=query,num_queries=3,llm=self._llm)
        with OptimizedMultiQuerySearcher(max_workers=4) as searcher:
            results = asyncio.run(searcher.search_multiple_queries(queries=rq,
                                                                   search_provider="google"))
        all_urls = [res.urls for res in results]
        non_empty_urls = [x for x in all_urls if x]
        filtered_urls = [y for x in zip(*non_empty_urls) for y in x]

        parser = FastHTMLParserV3(urls=filtered_urls[:10])
        content = asyncio.run(parser.fetch_content())
        content = [TextFilter.filter_text(c) for c in content[:3] if c]

        return "\n".join(content)


    # Function to create a task based on the user's question
    def _create_search_task(self,question):

        return Task(
            description=f"Research the following question: {question}",
            expected_output="A single line answer to the question",
            agent=self._researcher
        )

    # Function to run the crew and get results
    def run_question_answering(self,question):
        task = self._create_search_task(question)
        crew = Crew(
            agents=[self._researcher],
            tasks=[task]
        )
        result = crew.kickoff()
        return result

    def run_question_answering_loop(self):
        while True:
            user_question = input("Enter your question: \n")
            results = self.run_question_answering(user_question)
            print('#'*40)
            print(results)
            print('-'*40)

    def _setup_tool_and_search(self):
        self._search_tool = Tool(
            name="Google Search",
            func=self._web_search,
            description="Useful for when you need to answer questions about current events or the current state of the world"
        )
        self._researcher = Agent(
            role='Researcher',
            goal='Find accurate and up-to-date information',
            backstory='You are an AI research assistant skilled in finding information on the internet.',
            tools=[self._search_tool],
            llm=self._llm,
            verbose = True
        )


# Example usage
if __name__ == "__main__":
    llm = CustomChatOllama(
        model="qwen2:7b-instruct",
        base_url="http://192.168.162.49:8888",
        temperature=0
    )
    crew = CrewAIWebSearchQASystem(llm=llm)
    print(crew.run_question_answering(question="Latest langchain new version improvements?"))




import asyncio
import os
from functools import lru_cache

import aiohttp
import trafilatura
from bs4 import BeautifulSoup

from nltk.tokenize import sent_tokenize
from transformers import pipeline
import re

from sumy.parsers.plaintext import PlaintextParser
from sumy.nlp.tokenizers import Tokenizer
from sumy.summarizers.lsa import LsaSummarizer
from sumy.nlp.stemmers import Stemmer
from sumy.utils import get_stop_words
from .text_filter import TextFilter



class FastHTMLParser:
    def __init__(self, urls,fast_response = False):
        self.urls = urls
        self.fast_response = fast_response

    async def _fetch_url(self, session, url):
        try:
            async with session.get(url, timeout=10) as response:
                if response.status == 200:
                    html = await response.text()
                    return trafilatura.extract(html) or ""
                else:
                    return ""
        except Exception:
            return ""

    async def __fetch_all_urls(self):
        async with aiohttp.ClientSession() as session:
            if self.fast_response:
                for url in self.urls:
                    result = await self._fetch_url(session, url)
                    if result:
                        return [result]
                return []
            else:
                tasks = [self._fetch_url(session, url) for url in self.urls]
                return await asyncio.gather(*tasks)
    def fetch_content(self):
        return asyncio.run(self.__fetch_all_urls())


class FastHTMLParserV2:
    def __init__(self, urls, fast_response=False, max_length=1000, summarize=False):
        self.urls = urls
        self.fast_response = fast_response
        self.max_length = max_length
        self.summarize = summarize
        if summarize:
            self.summarizer = pipeline("summarization", model="facebook/bart-large-cnn")

    async def _fetch_url(self, session, url):
        try:
            async with session.get(url, timeout=10) as response:
                if response.status == 200:
                    html = await response.text()
                    return self._process_html(html)
                else:
                    return ""
        except Exception as e:
            # print(f"Error fetching {url}: {str(e)}")
            return ""

    def _process_html(self, html):
        # Extract main content using trafilatura
        extracted_text = trafilatura.extract(html) or ""

        # Use BeautifulSoup for additional cleaning
        soup = BeautifulSoup(html, 'html.parser')

        # Remove script, style, and nav elements
        for element in soup(["script", "style", "nav", "footer", "header"]):
            element.decompose()

        # Get text and clean it
        text = soup.get_text()
        lines = (line.strip() for line in text.splitlines())
        chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
        text = '\n'.join(chunk for chunk in chunks if chunk)

        # Combine trafilatura and BeautifulSoup results
        combined_text = f"{extracted_text}\n\n{text}"

        # Remove extra whitespace and newlines
        cleaned_text = re.sub(r'\s+', ' ', combined_text).strip()

        # Split into sentences
        sentences = sent_tokenize(cleaned_text)

        # Truncate if necessary
        if len(sentences) > self.max_length:
            sentences = sentences[:self.max_length]

        # Join sentences with newlines for better formatting
        formatted_text = '\n'.join(sentences)

        if self.summarize:
            return self._summarize_text(formatted_text)
        else:
            return formatted_text

    def _summarize_text(self, text):
        # Split text into chunks of 1000 tokens (approximate)
        chunks = [text[i:i + 4000] for i in range(0, len(text), 4000)]
        summaries = []
        for chunk in chunks:
            summary = self.summarizer(chunk, max_length=100, min_length=30, do_sample=False)
            summaries.append(summary[0]['summary_text'])
        return ' '.join(summaries)

    async def __fetch_all_urls(self):
        async with aiohttp.ClientSession() as session:
            if self.fast_response:
                for url in self.urls:
                    result = await self._fetch_url(session, url)
                    if result:
                        return [result]
                return []
            else:
                tasks = [self._fetch_url(session, url) for url in self.urls]
                return await asyncio.gather(*tasks)

    def fetch_content(self):
        return asyncio.run(self.__fetch_all_urls())

class FastHTMLParserV3:
    def __init__(self, urls, fast_response=False, max_length=None, summarize=False, summary_sentences=3):
        self.urls = urls
        self.fast_response = fast_response
        self.max_length = max_length
        self.summarize = summarize
        self.summary_sentences = summary_sentences
        if summarize:
            self.summarizer = LsaSummarizer(Stemmer('english'))
            self.summarizer.stop_words = get_stop_words('english')

    @lru_cache
    async def _fetch_url(self, session, url, url_fetch_timeout=10):
        if self._is_avoid_urls(url):
            return ""
        try:
            async with session.get(url, timeout=url_fetch_timeout) as response:
                if response.status == 200:
                    html = await response.text()
                    # return self._process_html(html)
                    return self._process_trafili_html(html)
                else:
                    # print(f"Error fetching {url}: HTTP status {response.status}")
                    return ""
        except aiohttp.ClientConnectorError as e:
            print(f"Connection error for {url}: {str(e)}")
            return ""
        except asyncio.TimeoutError:
            print(f"Timeout error for {url}")
            return ""
        except Exception as e:
            print(f"Unexpected error fetching {url}: {str(e)}")
            return ""

    def _process_html(self, html):
        try:
            soup = BeautifulSoup(html, 'html.parser')
            for element in soup(['script', 'style', 'nav', 'footer', 'header']):
                element.decompose()
            text = soup.get_text(separator=' ', strip=True)
            text = re.sub(r'\s+', ' ', text)
            text = re.sub(r'\[.*?\]', '', text)
            if self.max_length is None:
                self.max_length = len(text)

            if len(text) > self.max_length:
                text = text[:self.max_length]
            if self.summarize:
                return self._summarize_text(text)
            else:
                return text
        except Exception as e:
            print(f"Error processing HTML: {str(e)}")
            return ""

    def _summarize_text(self, text):
        try:
            parser = PlaintextParser.from_string(text, Tokenizer("english"))
            summary = self.summarizer(parser.document, self.summary_sentences)
            return ' '.join(str(sentence) for sentence in summary)
        except Exception as e:
            print(f"Error summarizing text: {str(e)}")
            return text

    async def fetch_content(self,url_fetch_timeout=10):
        async with aiohttp.ClientSession() as session:
            if self.fast_response:
                for url in self.urls:
                    result = await self._fetch_url(session, url,url_fetch_timeout)
                    if result:
                        return [result]
                return []
            else:
                tasks = [self._fetch_url(session, url,url_fetch_timeout) for url in self.urls]
                return await asyncio.gather(*tasks)
                # return [result for result in await asyncio.gather(*tasks) if result]


    def _is_avoid_urls(self,url):
        if "https://arxiv.org/pdf" in url.lower() or url.endswith(".pdf") or "youtube.com/watch" in url.lower():
            return True
        return False

    def _process_trafili_html(self, html):
        text = trafilatura.extract(html, include_formatting=True)
        text = TextFilter.filter_text(text)
        return text

#
# class FastHTMLParserV3:
#     def __init__(self, urls, fast_response=False, max_length=1000, summarize=False, summary_sentences=3):
#         self.urls = urls
#         self.fast_response = fast_response
#         self.max_length = max_length
#         self.summarize = summarize
#         self.summary_sentences = summary_sentences
#         if summarize:
#             self.summarizer = LsaSummarizer(Stemmer('english'))
#             self.summarizer.stop_words = get_stop_words('english')
#
#     async def _fetch_url(self, session, url):
#         try:
#             async with session.get(url, timeout=10) as response:
#                 if response.status == 200:
#                     html = await response.text()
#                     return self._process_html(html)
#                 else:
#                     return ""
#         except Exception as e:
#             print(f"Error fetching {url}: {str(e)}")
#             return ""
#
#     def _process_html(self, html):
#         # Use BeautifulSoup for content extraction and cleaning
#         soup = BeautifulSoup(html, 'lxml')
#
#         # Remove unwanted elements
#         for element in soup(['script', 'style', 'nav', 'footer', 'header']):
#             element.decompose()
#
#         # Extract text
#         text = soup.get_text(separator=' ', strip=True)
#
#         # Clean text
#         text = re.sub(r'\s+', ' ', text)
#         text = re.sub(r'\[.*?\]', '', text)  # Remove content in square brackets
#
#         # Truncate if necessary
#         if len(text) > self.max_length:
#             text = text[:self.max_length]
#
#         if self.summarize:
#             return self._summarize_text(text)
#         else:
#             return text
#
#     def _summarize_text(self, text):
#         parser = PlaintextParser.from_string(text, Tokenizer("english"))
#         summary = self.summarizer(parser.document, self.summary_sentences)
#         return ' '.join(str(sentence) for sentence in summary)
#
#     async def __fetch_all_urls(self):
#         async with aiohttp.ClientSession() as session:
#             if self.fast_response:
#                 for url in self.urls:
#                     result = await self._fetch_url(session, url)
#                     if result:
#                         return [result]
#                 return []
#             else:
#                 tasks = [self._fetch_url(session, url) for url in self.urls]
#                 return await asyncio.gather(*tasks)
#
#     async def fetch_content(self):
#         return await self.__fetch_all_urls()


from abc import ABC, abstractmethod
from typing import List


# L - Liskov Substitution Principle: Derived classes should be replaceable by their base class (for rewards, searchers, etc.)
class Searcher(ABC):
    search_provider = "google"
    @abstractmethod
    def perform_search(self, query: str) -> List[str]:
        pass

    @abstractmethod
    def perform_batch_search(self, batch_queries: List[str]) -> List[str]:
        pass




from fastembed import TextEmbedding
import numpy as np
from typing import List, Tuple
from langchain.text_splitter import RecursiveCharacterTextSplitter

from fastembed import TextEmbedding
import numpy as np
from typing import List, Tuple
from langchain.text_splitter import RecursiveCharacterTextSplitter
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.preprocessing import normalize


def fast_embedding_search(text_corpus: str, query: str, top_k: int = 5, top_tfidf=10,
                          chunk_size=500, chunk_overlap=50) -> List[Tuple[str, float]]:
    global embedding_model, tfidf_vectorizer

    # Initialize models (only once)
    if 'embedding_model' not in globals():
        embedding_model = TextEmbedding()
    if 'tfidf_vectorizer' not in globals():
        tfidf_vectorizer = TfidfVectorizer(lowercase=True, stop_words='english')

    text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
        separators=["\n\n", "\n", ".", "!", "?", ",", " ", ""]
    )

    # Split the corpus into sentences
    sentences = text_splitter.split_text(text_corpus)
    sentences = [sentence.strip() for sentence in sentences if sentence.strip()]

    # TF-IDF vectorization for initial filtering
    tfidf_matrix = tfidf_vectorizer.fit_transform(sentences)
    query_tfidf = tfidf_vectorizer.transform([query])

    # Calculate TF-IDF similarities
    tfidf_similarities = cosine_similarity(query_tfidf, tfidf_matrix).flatten()

    # Select top 2*top_k candidates based on TF-IDF similarity
    candidate_indices = np.argsort(tfidf_similarities)[-top_tfidf * top_k:][::-1]
    candidate_sentences = [sentences[i] for i in candidate_indices]

    # Batch encode the candidate sentences and query
    candidate_embeddings = np.array(list(embedding_model.embed(candidate_sentences)))
    query_embedding = np.array(list(embedding_model.embed([query])))[0]

    # Normalize embeddings for faster cosine similarity computation
    candidate_embeddings = normalize(candidate_embeddings)
    query_embedding = normalize(query_embedding.reshape(1, -1))

    # Compute cosine similarities
    cosine_scores = cosine_similarity(query_embedding, candidate_embeddings).flatten()

    # Get the top-k results
    top_indices = np.argsort(cosine_scores)[-top_k:][::-1]
    top_results = [(candidate_sentences[i], cosine_scores[i]) for i in top_indices]

    return top_results


def fast_embed_backup(text_corpus: str, query: str, top_k: int = 1) -> List[Tuple[str, float]]:
    # Initialize the TextEmbedding model
    embedding_model = TextEmbedding()

    text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
        chunk_size=500,
        chunk_overlap=50,
        separators=["\n\n", "\n", ".", "!", "?", ",", " ", ""]
    )

    # Split the corpus into sentences
    sentences = text_splitter.split_text(text_corpus)

    sentences = [sentence.strip() for sentence in sentences if sentence.strip()]

    # Encode the sentences and the query
    sentence_embeddings = list(embedding_model.embed(sentences))
    query_embedding = list(embedding_model.embed([query]))[0]

    # Convert to numpy arrays for efficient computation
    sentence_embeddings = np.array(sentence_embeddings)
    query_embedding = np.array(query_embedding)

    # Compute cosine similarities
    cosine_scores = np.dot(sentence_embeddings, query_embedding) / (
            np.linalg.norm(sentence_embeddings, axis=1) * np.linalg.norm(query_embedding)
    )

    # Get the top-k results
    top_results = sorted(zip(sentences, cosine_scores), key=lambda x: x[1], reverse=True)[:top_k]

    return top_results


import asyncio
import time
from typing import List
from langchain_core.language_models import BaseChatModel
from langchain.prompts import PromptTemplate
from langchain_core.output_parsers import StrOutputParser

from .agent import WebSearchAgent, OptimizedMultiQuerySearcher
from .html_parser import FastHTMLParserV3
from .text_filter import TextFilter
from ..langchain_ollama import CustomChatOllama


class CrewAIWebSearchQASystem:
    def __init__(self, llm: BaseChatModel):
        self._llm = llm
        self._web_agent = WebSearchAgent()
        self._setup_chain()

    def _setup_chain(self):
        prompt = PromptTemplate(
            input_variables=["question", "search_results"],
            template="""
            You are an AI research assistant. Use the following search results to answer the question.
            Question: {question}
            Search Results: {search_results}
            Please provide a concise and accurate answer based on the search results.
            """
        )
        self._chain = prompt | self._llm | StrOutputParser()

    async def _web_search(self, query: str) -> str:
        """Run query through GoogleSearch and parse result."""
        refined_queries = self._web_agent.refine_queries(user_query=query, num_queries=3, llm=self._llm)
        searcher = OptimizedMultiQuerySearcher(max_workers=4)
        results = await searcher.search_multiple_queries(
            queries=refined_queries
        )

        all_urls = [res.urls for res in results]
        non_empty_urls = [x for x in all_urls if x]
        filtered_urls = [y for x in zip(*non_empty_urls) for y in x]

        parser = FastHTMLParserV3(urls=filtered_urls[:10])
        content = await parser.fetch_content()
        filtered_content = [TextFilter.filter_text(c) for c in content[:3] if c]
        return "\n".join(filtered_content)

    async def answer_question(self, question: str) -> str:
        search_results = await self._web_search(question)

        response = await self._chain.ainvoke({
            "question": question,
            "search_results": search_results
        })

        return response.strip()

    async def run_question_answering_loop(self):
        while True:
            user_question = input("Enter your question: \n")
            answer = await self.answer_question(user_question)
            print('#' * 40)
            print(answer)
            print('-' * 40)


# Example usage
if __name__ == "__main__":
    llm = CustomChatOllama(
        model="qwen2:7b-instruct",
        base_url="http://192.168.162.49:8888",
        temperature=0
    )
    qa_system = CrewAIWebSearchQASystem(llm=llm)

    question = "Latest langchain new version improvements?"
    answer = asyncio.run(qa_system.answer_question(question))
    print(answer)

    # Uncomment to run in loop mode
    # asyncio.run(qa_system.run_question_answering_loop())


import json
import os
import logging
import networkx as nx
import matplotlib.pyplot as plt

from datetime import datetime
from typing import List
from .researcher import AiResearcher
from .semantic_chunker import SemanticSimilarityTextChunker
from .llm_service.base import LLMPrompt
from .prompts import Prompt

class MultiAgentQueryOrchestrator:
    def __init__(self,response_level="medium",device = "Local",model = None):
        self.device = device
        self.results = {}
        self.response_level = response_level
        self.results_folder = False
        self.researcher = AiResearcher()
        self.setup_logging()

        self.num_queries = 3
        self.model = model
        self.subquery_max_urls = 3
        self.subquery_max_articles = 3
        self.top_chunks_in_article = 3
        self.max_subquery_relevant_chunks = 5
        self.sentence_chunk_size_in_article = 3
        self.setup_args()

    def setup_logging(self):
        logging.basicConfig(level=logging.INFO,
                            format='%(asctime)s - %(levelname)s - %(message)s',
                            handlers=[logging.StreamHandler()])
        self.logger = logging.getLogger(__name__)

    def create_results_folder(self):
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        folder_name = f"Results/results_{timestamp}"
        os.makedirs(folder_name, exist_ok=True)
        return folder_name

    def save_to_results(self, filename, content):
        if self.results_folder:
            full_path = os.path.join(self.results_folder, filename)
            with open(full_path, 'w') as f:
                if isinstance(content, str):
                    f.write(content)
                else:
                    json.dump(content, f, indent=2)
        else:
            file_name= filename.rsplit('.', 1)[0]
            self.results.update({file_name:content})
        self.logger.info(f"Saved {filename} to results folder")

    def decompose_query(self, query: str, num_queries: int) -> List[str]:
        self.logger.info(f"Decomposing query: {query}")
        decompose_prompt = LLMPrompt(Prompt.DECOMPOSE_PROMPT,model=self.model,
                                     device=self.device
                                     )
        subqueries_str = decompose_prompt.run(websearchquery=query,
                                              num_queries=num_queries,
                                              return_type=list)
        subqueries_str = subqueries_str.replace("'", '"')
        subqueries = json.loads(subqueries_str)
        self.save_to_results("subqueries.json", subqueries)
        return subqueries

    def visualize_dependency_graph(self, G: nx.DiGraph):
        self.logger.info("Visualizing dependency graph")
        plt.figure(figsize=(12, 8))
        pos = nx.spring_layout(G)
        nx.draw(G, pos, with_labels=True, node_color='lightblue',
                node_size=2000, font_size=10, arrows=True)

        labels = nx.get_node_attributes(G, 'query')
        label_pos = {k: (v[0], v[1] + 0.02) for k, v in pos.items()}
        nx.draw_networkx_labels(G, label_pos, labels, font_size=8, font_weight="bold")

        edge_labels = {(u, v): "" for (u, v) in G.edges()}
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, font_size=8)

        plt.title("Subquery Dependency Graph")
        plt.axis('off')
        plt.tight_layout()
        filename = os.path.join(self.results_folder, "dependency_graph.png")
        plt.savefig(filename, format="png", dpi=300, bbox_inches="tight")
        plt.close()
        self.logger.info(f"Graph visualization saved as {filename}")

    def create_dependency_graph(self, subqueries: List[str]) -> nx.DiGraph:
        self.logger.info("Creating dependency graph")
        G = nx.DiGraph()
        for i, sq in enumerate(subqueries):
            G.add_node(i, query=sq)

        for i in range(len(subqueries)):
            for j in range(i + 1, len(subqueries)):
                dependency_prompt = LLMPrompt(Prompt.NODE_DEPENDENCY_PROMPT,model=self.model,
                                              device=self.device
                                              )
                dependency = dependency_prompt.run(first=subqueries[i],
                                                   second=subqueries[j]).strip().lower()
                if 'yes' in dependency:
                    G.add_edge(j, i)

        return G

    def process_subquery(self, subquery: str, context: str = "") -> str:
        self.logger.info(f"Processing subquery: {subquery}")
        articles = self.researcher.get_query_content(subquery, max_articles=self.subquery_max_articles,
                                                     max_urls=self.subquery_max_urls)
        chunker = SemanticSimilarityTextChunker(subquery)
        all_chunks = []
        for article in articles:
            chunks = chunker.chunk_text(article,
                                        top_chunks_in_article=self.top_chunks_in_article,
                                        chunk_size=self.sentence_chunk_size_in_article)
            all_chunks.extend(chunks)
        all_chunks.sort(key=lambda chunk: chunker.score_sentence(chunk, chunker.get_query_keywords()), reverse=True)
        relevant_chunks = all_chunks[:self.max_subquery_relevant_chunks]  # Top 3 chunks

        combined_input = context + "\n\n" + "\n".join(relevant_chunks)
        answer_prompt = LLMPrompt(Prompt.ANSWER_PROMPT,model=self.model,
                                  device=self.device
                                  )
        return answer_prompt.run(combined_input=combined_input, subquery=subquery)

    def save_graph(self, G: nx.DiGraph):
        self.logger.info("Saving dependency graph")
        data = nx.node_link_data(G)
        self.save_to_results("dependency_graph.json", data)

    def load_graph(self) -> nx.DiGraph:
        self.logger.info("Loading dependency graph")
        filename = os.path.join(self.results_folder, "dependency_graph.json")
        with open(filename, 'r') as f:
            data = json.load(f)
        return nx.node_link_graph(data)

    def extract_relevant_info(self, subquery: str, web_result: str, child_context: str) -> str:
        self.logger.info(f"Extracting relevant info for subquery: {subquery}")
        runner = LLMPrompt(Prompt.RELEVANT_INFO_PROMPT,model=self.model,
                           device=self.device
                           )
        return runner.run(
            subquery=subquery,
            web_result=web_result,
            child_context=child_context
        )

    def process_query(self, query: str, num_queries=None) -> str:
        if num_queries is None:
            num_queries = self.num_queries
        self.logger.info(f"Processing main query: {query}")
        subqueries = self.decompose_query(query=query, num_queries=num_queries)
        self.logger.info(f"Subqueries\n {subqueries}")
        graph = self.create_dependency_graph(subqueries)
        if self.results_folder:
            self.visualize_dependency_graph(graph)
            self.save_graph(graph)

        results = {}
        for node in reversed(list(nx.topological_sort(graph))):
            results[node] = {}
            subquery = graph.nodes[node]['query']
            self.logger.info(f"Processing subquery: {subquery}")
            web_result = self.process_subquery(subquery)
            results[node]["web_result"] = web_result

            child_results = [results.get(child, {}).get("relevant_info", "") for child in graph.successors(node)]
            child_context = "\n".join(child_results)

            relevant_info = self.extract_relevant_info(subquery, web_result, child_context)
            results[node]["relevant_info"] = relevant_info

        self.save_to_results("subquery_results.json", results)

        final_context = "\n".join([f"Subquery {i + 1}: {subquery}\nResult: {results[i]['relevant_info']}\n"
                                   for i, subquery in enumerate(subqueries)])

        self.logger.info("Generating final answer")
        runner = LLMPrompt(Prompt.COMPREHENSIVE_ANSWER_PROMPT,model=self.model,
                           device=self.device
                           )
        final_answer = runner.run(
            final_context=final_context,
            query=query
        )
        self.save_to_results("final_answer.txt", final_answer)
        return final_answer

    def setup_args(self):
        if self.results_folder:
            self.results_folder = self.create_results_folder()

        if self.response_level=="low":
            self.num_queries=3
            self.model = "qwen2:1.5b-instruct" if self.model is None else self.model
            self.subquery_max_urls = 2
            self.subquery_max_articles = 2
            self.top_chunks_in_article = 2
            self.max_subquery_relevant_chunks = 2
            self.sentence_chunk_size_in_article = 2
        elif self.response_level=="high":
            self.num_queries=8
            self.model = "qwen2:7b-instruct" if self.model is None else self.model
            self.subquery_max_urls = 4
            self.subquery_max_articles = 4
            self.top_chunks_in_article = 4
            self.max_subquery_relevant_chunks = 5
            self.sentence_chunk_size_in_article = 3
        else:
            self.num_queries=5
            self.model = "qwen2:7b-instruct" if self.model is None else self.model
            self.subquery_max_urls = 3
            self.subquery_max_articles = 3
            self.top_chunks_in_article = 3
            self.max_subquery_relevant_chunks = 5
            self.sentence_chunk_size_in_article = 3


if __name__ == '__main__':
    processor = MultiAgentQueryOrchestrator(
            response_level="low",
            device="GPU"
        )
    query = "when is pawan kalyan new movie OG release date?"
    result = processor.process_query(query)
    print(processor.results)


from .embed_search import fast_embedding_search
from .html_parser import FastHTMLParserV3
from .optimized_multi_query_searcher import OptimizedMultiQuerySearcher
from .agent import WebSearchAgent,WebSearchArguments

from .crewai_web_serach_qa import CrewAIWebSearchQASystem
__all__ = ['fast_embedding_search', 'WebSearchAgent', 'FastHTMLParserV3', 'WebSearchArguments',
           'OptimizedMultiQuerySearcher',
           "CrewAIWebSearchQASystem"
           ]


import asyncio
from fastapi import FastAPI
from fastapi.responses import StreamingResponse, JSONResponse
from langchain_ollama import ChatOllama
from langchain.schema import HumanMessage
from pydantic import BaseModel
from typing import List

app = FastAPI()

class ChatInput(BaseModel):
    text: str
    model: str
    temperature: float
    num_predict: float


class BatchChatInput(BaseModel):
    messages: List[str]
    model: str
    temperature: float

async def chat_stream(text: str, model: str, temp: float,num_predict):
    chat_model = ChatOllama(model=model, temperature=temp,num_predict = num_predict)
    async for chunk in chat_model.astream([HumanMessage(content=text)]):
        yield chunk.content

async def process_batch(messages: List[str], model: str, temp: float):
    chat_model = ChatOllama(model=model, temperature=temp)
    # Convert messages to HumanMessage objects
    human_messages = [HumanMessage(content=msg) for msg in messages]
    # Use the batch method for efficient processing
    responses = await chat_model.abatch(human_messages)
    # Extract content from responses
    return [response.content for response in responses]

@app.post("/chat")
async def chat_endpoint(chat_input: ChatInput):
    return StreamingResponse(chat_stream(chat_input.text, chat_input.model, chat_input.temperature,chat_input.num_predict), media_type="text/event-stream")

@app.post("/batch-chat")
async def batch_chat_endpoint(batch_input: BatchChatInput):
    results = await process_batch(batch_input.messages, batch_input.model, batch_input.temperature)
    return JSONResponse(content={"results": results})

if __name__ == "__main__":
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8888)





from researcher_ai.main.parse_url.html_parser import UrlTextParser


def test_url_text_parser():
    parser = UrlTextParser()

    # Test with a PDF URL
    pdf_urls = ["https://arxiv.org/pdf/2402.15000"]
    pdf_result = parser.parse_html(pdf_urls)
    assert len(pdf_result) == 1, "Expected 1 result for PDF URL"

def test_html_parser():
    parser = UrlTextParser()
    # Test with an HTML URL
    html_urls = ["https://python.langchain.com/v0.2/docs/introduction/"]
    html_result = parser.parse_html(html_urls)
    assert len(html_result) == 1, "Expected 1 result for HTML URL"





import re


class TextFilter:
    @staticmethod
    def _remove_mulitline_space(text):
        text = text.replace("\n\n", "\n")
        return text

    @staticmethod
    def _replace_single_line_words_to_single_sentence(text):
        for match in re.findall(r'\w+\n\w+\n', text):
            text = text.replace(match, match.replace('\n', " "))
        return text

    @staticmethod
    def filter_text(text):
        text = TextFilter._remove_mulitline_space(text)
        text = TextFilter._replace_single_line_words_to_single_sentence(text)
        text = TextFilter._remove_mulitline_space(text)
        return text


from ..semantic_chunker import SemanticSimilarityTextChunker


class ContextExtractor:
    def __init__(self, top_chunks_in_article=3, sentence_chunk_size_in_article=3, max_subquery_relevant_chunks=3):
        self.top_chunks_in_article = top_chunks_in_article
        self.sentence_chunk_size_in_article = sentence_chunk_size_in_article
        self.max_subquery_relevant_chunks = max_subquery_relevant_chunks

    def extract_context(self, query, articles):
        chunker = SemanticSimilarityTextChunker(query)
        all_chunks = []

        for article in articles:
            chunks = chunker.chunk_text(
                article,
                top_chunks_in_article=self.top_chunks_in_article,
                chunk_size=self.sentence_chunk_size_in_article
            )
            all_chunks.extend(chunks)

        all_chunks.sort(
            key=lambda chunk: chunker.score_sentence(chunk, chunker.get_query_keywords()),
            reverse=True
        )

        relevant_chunks = all_chunks[:self.max_subquery_relevant_chunks]
        return relevant_chunks

# Example usage:
# extractor = ContextExtractor(top_chunks_in_article=5, sentence_chunk_size_in_article=3, max_subquery_relevant_chunks=10)
# relevant_chunks = extractor.extract_context("example query", ["article1 text", "article2 text"])


import asyncio
import re
from typing import List

from .base import Searcher


class GoogleSearch(Searcher):
    def perform_search(self, query: str) -> List[str]:
        # Here you'd implement the actual API search using google_searcher
        print(f"Searching Google for: {query}")
        return [f"Result {i + 1} for {query}" for i in range(5)]  # Dummy results

    def perform_batch_search(self, batch_queries: List[str]) -> List[str]:
        pass

import asyncio
from typing import List
from opengenai.web_search import OptimizedMultiQuerySearcher

class RealTimeGoogleSearchProvider(Searcher):
    def __init__(self, search_provider="google"):
        self.search_provider = search_provider
        self.searcher = OptimizedMultiQuerySearcher()

    def perform_search(self, query: str,max_urls=5) -> List[str]:
        return self.searcher.search_single_query(query,search_provider=self.search_provider).urls[:max_urls]


    async def _async_batch_search(self, batch_queries,max_urls=5) -> List[str]:
        all_urls = await self.searcher.search_multiple_queries(
            queries=batch_queries,
            search_provider=self.search_provider
        )
        all_urls = [url.urls for url in all_urls]
        filtered_urls = [y for x in zip(*all_urls) for y in x]
        filtered_urls = [self.extract_until_hash(x) if self.is_hash(x) else x for x in filtered_urls]
        filtered_urls = [_ for _ in filtered_urls if _]
        return filtered_urls[:max_urls]

    def perform_batch_search(self, batch_queries,max_urls=5) -> List[str]:
        return asyncio.run(self._async_batch_search(batch_queries,max_urls=max_urls))

    def is_hash(self, x):
        return '#' in x

    def extract_until_hash(self, x):
        results = re.findall(r'(.*)#',x)
        if results:
            return results[0]
        return ""


import asyncio
import re
from io import BytesIO
from typing import List, Dict, Union, Any
from datetime import datetime
import json
import logging

from langchain_text_splitters import RecursiveCharacterTextSplitter
from pydantic import BaseModel, Field
from weasyprint import HTML
import markdown
from langchain.prompts import PromptTemplate
from langchain_ollama import ChatOllama
from langchain.schema import StrOutputParser
from .optimized_multi_query_searcher import OptimizedMultiQuerySearcher

# Set up logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)


class WebSearchHelper:
    @staticmethod
    def clean_query(query: str, query_with_date: bool = False) -> str:
        if not query:
            return query
        query = re.sub(r'[^a-zA-Z0-9\s]', '', query)

        if query_with_date:
            today_date = ", " + datetime.now().strftime('%B-%d-%Y')
        else:
            today_date = ""

        query = f"{query}{today_date}"
        return ' '.join(query.split())

    @staticmethod
    def summarize_text(llm, text: str) -> str:
        text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
            chunk_size=250,
            chunk_overlap=25,
            is_separator_regex=False,
        )
        chunks = text_splitter.split_text(text)

        summarize_prompt = PromptTemplate.from_template(
            """Summarize the following [CHUNK], focusing on the main points:
            - Maintain key terms and concepts
            - Keep the language natural and easy to understand
            - Do not introduce new information, don't say this text, directly start summary

            [CHUNK]
            {chunk}

            Summary:"""
        )

        summarize_chain = llm | StrOutputParser()
        prompts = [summarize_prompt.format(chunk=chunk) for chunk in chunks]
        summaries: List[str] = summarize_chain.batch(prompts)

        combined_summary = "\n".join(summaries)
        combined_summary = f"### Summary: \n{combined_summary}\n\n"
        combined_summary += f"### Original result: \n{text}\n\n---\n\n"
        return combined_summary

    @staticmethod
    def get_md_content(result,summarize = False):
        md_content = f"# Search Results for: \n{result['user_query']}\n\n"
        md_content += f"## Refined Queries\n\n"
        for query in result['refined_queries']:
            md_content += f"- {query}\n"
        md_content += f"\n## Search Results\n\n"

        for i, search_result in enumerate(result['search_results'], 1):
            if summarize and "no valid content" not in search_result.lower():
                md_content += WebSearchHelper.summarize_text(llm, search_result)
            else:
                md_content += f"### Result {i}\n\n{search_result['results']}\n\n---\n\n"

        md_content += f"\n## URLs\n\n"
        for i, url_list in enumerate(result['search_results'], 1):
            try:
                _query = result['refined_queries'][i - 1]
            except:
                _query = f"Query {i}"
            md_content += f"### {_query}\n\n"
            for url in url_list['urls']:
                md_content += f"- [{url}]({url})\n"
        return md_content

    @staticmethod
    def save_results(result: Dict, base_filename: str = "search_results", summarize: bool = False, llm=None) -> str:
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        filename = f"{base_filename}_{timestamp}"
        # os.makedirs("Results", exist_ok=True)
        # with open(f"Results/{filename}.json", "w") as f:
        #     json.dump(result, f, indent=2)

        md_content = WebSearchHelper.get_md_content(result=result,
                                                    summarize=summarize)



        return md_content

    @staticmethod
    def generate_pdf(md_content):
        html = markdown.markdown(md_content)

        # Add CSS for better URL display
        html = f"""
        <style>
        .url-list {{
            margin-top: 20px;
            padding: 10px;
            background-color: #f0f0f0;
            border-radius: 5px;
        }}
        .url-list a {{
            display: block;
            margin-bottom: 5px;
            word-break: break-all;
        }}
        </style>
        {html}
        """

        # Create a BytesIO object to store the PDF
        pdf_buffer = BytesIO()

        # Generate the PDF and write it to the BytesIO buffer
        HTML(string=html).write_pdf(pdf_buffer)

        # Get the PDF content as bytes
        pdf_bytes = pdf_buffer.getvalue()

        return pdf_bytes

    @staticmethod
    def get_md_content_v2(search_results,user_query,refined_queries):
        md_content = f"# Search Results for: \n{user_query}\n\n"
        md_content += f"## Refined Queries\n\n"
        for query in refined_queries:
            md_content += f"- {query}\n"
        md_content += f"\n## Search Results\n\n"
        for idx,x in enumerate(search_results):
            md_content += f"### Result{idx}:\n {x['context']}\n\n---\n\n"
        md_content += "## URLS:\n"
        for idx,x in enumerate(search_results,1):
            md_content += f"{idx}:{x['url']}\n\n"

        return md_content





class WebSearchAgentPersonas:
    expert_personas = [
        "100-year-old Data Science Guru",
        "Omniscient Web Search Wizard",
        "AI-powered Research Mastermind"
    ]

    @classmethod
    def get_query_refining_prompt(cls, num_queries: int):
        chosen_persona = " and ".join(cls.expert_personas)

        QUERY_REFINING_PROMPT = f"""You are the {chosen_persona}, a legendary search expert with unparalleled ability to find information on any topic. Your task is to generate {num_queries} highly effective search queries based on the following user query: {{user_query}}

        Guidelines for generating queries:
        1. Analyze the user's intent deeply and break it down into core concepts.
        2. Create queries that cover all aspects of the user's request, including potential implicit needs.
        3. Use advanced search operators when appropriate (e.g., site:github.com, filetype:pdf, intitle:, etc.).
        4. Include specific technical terms, frameworks, or methodologies relevant to the topic.
        5. Consider time-sensitive information by including  "latest" keywords.
        6. Combine broad exploratory queries with highly specific ones to ensure comprehensive coverage.
        7. Anticipate related topics or technologies that might yield valuable results.
        8. Craft queries as if you're an expert in the field, using insider knowledge and terminology.
        9. Ensure each query is unique and adds value beyond the others.
        10. Optimize for both relevance and recency of results.

        Format each query directly don't add extra terms as a separate line, without numbering or additional text."""

        # Format: Return a Python list of strings, each string being a refined search query. Do not include any explanations or numbering."""

        QUERY_REFINING_PROMPT_V2 = """You are an AI-powered search query optimizer. Your task is to analyze the given user query and generate {num_queries} refined search queries that will yield the most relevant and comprehensive results. 

Consider the following aspects when refining the query:
1. Core concepts and potential subtopics
2. Technical terminology and domain-specific language
3. Recent developments and time-sensitive information
4. Different perspectives or applications of the topic
5. Potential data sources or content types (e.g., academic papers, industry reports, code repositories)

For the user query: {user_query}

Generate diverse, effective search queries that capture the essence of the user's information need while expanding its scope to ensure comprehensive coverage. Each query should offer a unique angle or approach to finding relevant information.

Format each query directly as a separate line, each string being a refined search query. Do not include any explanations or numbering."""

        return QUERY_REFINING_PROMPT_V2


class WebSearchAgent:
    def __init__(self,query_refining_prompt=None):
        self.web_fetcher = None
        self.query_refining_prompt = query_refining_prompt

    def refine_queries(self, user_query: str, num_queries: int,llm) -> List[str]:
        if self.query_refining_prompt is None:
            self.query_refining_prompt = WebSearchAgentPersonas.get_query_refining_prompt(num_queries=num_queries)
            # self.query_refining_prompt = self.get_query_refining_prompt()
        query_refinement_prompt = PromptTemplate.from_template(self.query_refining_prompt)
        query_refinement_chain = query_refinement_prompt | llm | StrOutputParser() | (lambda x: x.split("\n"))
        refined_queries = query_refinement_chain.invoke({"user_query": user_query, "num_queries": num_queries})
        refined_queries = [x for x in refined_queries if len(x) > 0]
        return refined_queries

    async def search_web(self, refined_queries: List[str],
                         top_k=5, fast_response=False,
                         chromedriver_path="/usr/local/bin/chromedriver",
                         max_length=None, summarize=False, summary_sentences=3,
                         search_provider="google",
                         num_results=10,
                         url_fetch_timeout=10,
                         extract_url_content=False
                         ):
        logger.info("Searching the web")
        search_results = []


        for query in refined_queries:
            logger.info(f"Searching for: {query}")
            results, search_res = await self.web_fetcher.fetch(query,
                                                               top_k=top_k, fast_response=fast_response,
                                                               chromedriver_path=chromedriver_path,
                                                               max_length=max_length, summarize=summarize,
                                                               summary_sentences=summary_sentences,
                                                               search_provider=search_provider,
                                                               num_results=num_results,
                                                               url_fetch_timeout=url_fetch_timeout,
                                                               extract_url_content=extract_url_content
                                                               )
            if "no valid" in results.strip():
                logger.warning(f"Empty result for query: {query}")
            else:
                logger.info(f"Found results for query: {query}")

            search_results.append(dict(
                results=results,
                urls=search_res.urls,
                top_response=search_res.top_response,
                title_and_links=search_res.title_and_links,
                html_content=search_res.html_content
            ))
        logger.info(f"Completed web search for {len(search_results)} queries")
        return search_results

    async def arun(
            self,
            user_query: str,
            num_queries: Union[None,int] = None,
            llm = None,
            summarize: bool = False,
            summary_sentences: int = 3,
            query_with_date: bool = True,
            top_k: int = 5,
            fast_response: bool = False,
            chromedriver_path: str = "/usr/local/bin/chromedriver",
            max_length: Union[None, int] = None,
            search_provider=None,
            num_results=10,
            url_fetch_timeout=10,
            extract_url_content=False,
            return_markdown_text = True,
            return_json = False
    , return_pdf=False,
            return_only_urls  = False,
            max_workers=None) -> Union[Dict,str,bytes]:
        logger.info(f"Starting web search agent for query: {user_query}")
        if num_queries is not None and llm is None:
            raise ValueError("If num_queries is provided, llm should be provided as well.")
        if summarize is True:
            extract_url_content = True
        if num_queries is not None:
            refined_queries = self.refine_queries(user_query, num_queries,llm)
            refined_queries = [WebSearchHelper.clean_query(query, query_with_date) for query in refined_queries]
        else:
            refined_queries = [user_query]
        logger.info(f"Optimized queries: {refined_queries}")
        gs = OptimizedMultiQuerySearcher(chromedriver_path=chromedriver_path,
                                         max_workers=max_workers)
        all_urls = await gs.search_multiple_queries(refined_queries, search_provider=search_provider)
        # filtered_urls = [each_url for url in all_urls for each_url in url.urls if len(each_url)>0]
        all_urls = [url.urls for url in all_urls]
        filtered_urls = [y for x in zip(*all_urls) for y in x]
        if return_only_urls:
            return filtered_urls
        # res = await WebFetcher.get_url_content(filtered_urls)
        if extract_url_content:
            from opengenai.web_search.html_parser import FastHTMLParserV3
            parser = FastHTMLParserV3(urls=filtered_urls, fast_response=fast_response,
                                      max_length=max_length, summarize=summarize, summary_sentences=summary_sentences)
            contents = await parser.fetch_content(url_fetch_timeout=url_fetch_timeout)
        else:
            contents = [""]*len(filtered_urls)

        search_results = []
        for r,c in zip(filtered_urls,contents):
            search_results.append(dict(
                url=r,
                context = c
            ))

        # search_results = await self.search_web(refined_queries,
        #                                        top_k=top_k, fast_response=fast_response,
        #                                        chromedriver_path=chromedriver_path,
        #                                        max_length=max_length, summarize=summarize,
        #                                        summary_sentences=summary_sentences,
        #                                        search_provider=search_provider,
        #                                        num_results=num_results,
        #                                        url_fetch_timeout=url_fetch_timeout,
        #                                        extract_url_content=extract_url_content
        #                                        )
        md_content = WebSearchHelper.get_md_content_v2(search_results,
                                                       user_query,
                                                       refined_queries)
        # md_content = WebSearchHelper.save_results({
        #     "user_query": user_query,
        #     "refined_queries": refined_queries,
        #     "search_results": search_results
        # }, summarize=summarize, llm=self.llm)
        logger.info("Web search agent completed successfully")
        if return_pdf:
            return WebSearchHelper.generate_pdf(md_content=md_content)

        if return_json:
            return {
                "user_query": user_query,
                "refined_queries": refined_queries,
                "search_results": search_results,
                "md_content": md_content,
            }
        elif return_markdown_text:
            return md_content
        else:
            return md_content

    def get_query_refining_prompt(self):
        QUERY_REFINING_PROMPT = """You are GoogleSearch Expert , Generate {num_queries} different variations search queries based on the following user query: {user_query}

                    Guidelines for generating queries:
                    1. Generate in best form to get almost perfect result for user query
                    2. Add Meaningfull Search Terms to Enrich the Query
                    3. Avoid overly broad or vague queries

                    Format each query directly don't add extra terms as a separate line, without numbering or additional text."""
        return QUERY_REFINING_PROMPT


class WebSearchArguments(BaseModel):
    user_query: str
    num_queries: Union[int,None] = None
    summarize: bool = False
    summary_sentences: int = 3
    query_with_date: bool = True
    top_k: int = 3
    llm:Any = None
    fast_response: bool = False
    chromedriver_path: str = "/usr/local/bin/chromedriver"
    max_length: Union[int, None] = None
    search_provider: Union[None,str] = Field(
        default=None,
        description="Search engine to use for the search. Supported options: google, bing",
    )
    max_workers : int = 4
    num_results: int = 10
    url_fetch_timeout: int = 10
    extract_url_content: bool = False
    return_markdown_text: bool = True,
    return_json: bool = False
    return_pdf:bool = False


# Example usage
if __name__ == "__main__":
    # llm = ChatOllama(model="gemma2:2b")
    from opengenai.langchain_ollama.ollama import CustomChatOllama
    llm  = CustomChatOllama(
        model = "gemma2:2b",
        base_url = "http://192.168.162.49:8888"
    )
    agent = WebSearchAgent()
    args = WebSearchArguments(
        user_query="i want to know the latest agentic RAG based approaches",
        num_queries=3,
        query_with_date=False,
    )
    result = asyncio.run(agent.arun(**args.model_dump()))
    with open("res.md","w",encoding="utf-8") as f:
        f.write(result)



from langchain_ollama import ChatOllama
from opengenai.langchain_ollama import CustomChatOllama

database = {
    "list_prompt":"\n\nalways return python list directly, i will directly use your response to Json.parse() expecting List example like ['first','second'...]"
}

class LLMPrompt:
    def __init__(
            self,
            template: str,
            model: str = "qwen2:7b-instruct",
            base_url: str = "http://192.168.162.49:8888",
            temperature: float = 0.0,
            device = "GPU"
    ):
        self.template = template
        self.llm = CustomChatOllama(
            model=model,
            base_url=base_url,
            temperature=temperature
        ) if device=="GPU" else ChatOllama(model=model,temperature=temperature)

    def get_formatted_prompt(self,**variables):
        formatted_prompt = self.template.format(**variables)
        return formatted_prompt + (database.get("list_prompt") if variables.get("return_type") == list else "")

    def __call__(self, **variables):
        return self.run(**variables)

    def run(self,**variables):
        p = self.get_formatted_prompt(**variables)
        response = self.llm.invoke(p).content
        return response

    def stream(self, **variables):
        for response in self.llm.stream(self.get_formatted_prompt(**variables)):
            yield response.content


class Prompt:
    ANSWER_PROMPT = """Based on the following context and the question, provide a comprehensive  and only related content answer to the question:

                Context: {combined_input}

                Question: {subquery}
                """

    DECOMPOSE_PROMPT = """
                Task: Decompose the following user request websearchquery into maximum of {num_queries} distinct subwebsearchquery that together will comprehensively answer the main websearchquery.
                Note: there is no restriction to get the exact {num_queries} subwebsearchquery it can be less than max if it covers our user whole point of view.
                Main websearchquery: "{websearchquery}"

                Instructions:
                1. Each subwebsearchquery should focus on a specific aspect of the main websearchquery.
                2. Ensure the websearchquery are clear, concise, and directly related to the main subwebsearchquery.
                3. The subwebsearchquery should cover different aspects and not be repetitive.
                """

    NODE_DEPENDENCY_PROMPT = """Does the answer to '{first}' depend on the answer to '{second}'? 
                        Answer with 'Yes' or 'No'."""

    RELEVANT_INFO_PROMPT = """
            Based on the following information:

            Subquery: {subquery}
            Web search result: {web_result}
            Context from child queries: {child_context}

            Extract and summarize the most relevant information that answers the subquery and would be useful for parent queries. 
            Ensure the summary is concise but comprehensive.
            """

    COMPREHENSIVE_ANSWER_PROMPT = """
            Based on the following context, provide a comprehensive answer to the original query:

            Context:
            {final_context}

            Original Query: {query}

            Please synthesize the information from all subqueries to give a complete and coherent answer.
            """



import asyncio
from dataclasses import dataclass
import aiohttp
from io import BytesIO
import PyPDF2

from ....web_search import FastHTMLParserV3

from .base import BaseHtmlParser


@dataclass
class UrlTextParser(BaseHtmlParser):
    extract_pdf = True
    def parse_html(self, urls: list) -> list:
        return asyncio.run(self._async_html_parser(urls))

    async def _async_html_parser(self, urls):
        html_urls = []
        pdf_urls = []
        for url in urls:
            if '/pdf' in url or url.lower().endswith('.pdf'):
                pdf_urls.append(url)
            else:
                html_urls.append(url)

        results = []

        if html_urls:
            fetcher = FastHTMLParserV3(urls=html_urls)
            html_results = await fetcher.fetch_content()
            results.extend(html_results)

        if pdf_urls and self.extract_pdf:
            pdf_results = await self._fetch_pdf_content(pdf_urls)
            results.extend(pdf_results)

        return results

    async def _fetch_pdf_content(self, pdf_urls):
        async def fetch_pdf(session, url):
            async with session.get(url) as response:
                if response.status == 200:
                    pdf_content = await response.read()
                    pdf_file = BytesIO(pdf_content)
                    pdf_reader = PyPDF2.PdfReader(pdf_file)
                    text = ""
                    for page in pdf_reader.pages:
                        text += page.extract_text()
                    return text
                else:
                    return ""

        async with aiohttp.ClientSession() as session:
            tasks = [fetch_pdf(session, url) for url in pdf_urls]
            results = await asyncio.gather(*tasks)
            return results





from .parse_url.html_parser import UrlTextParser
from .search_provider.searcher import RealTimeGoogleSearchProvider

class AiResearcher:
    searcher = RealTimeGoogleSearchProvider()
    parser = UrlTextParser()

    def get_query_content(self,query,max_articles = 3,max_urls=6):
        """
        Gets the top 3 URLs for the given query, fetches their content,
        returns the first three contents.'
        """
        urls = self.searcher.perform_search(query,
                                       max_urls=max_urls)
        urls = list(set(urls))
        content = self.parser.parse_html(urls=urls)
        content = [x for x in content if x]
        return content[:max_articles]